
;       LOCATE.COM   v1.12   02-16-1996   C. Dye
;
;       LOCATE.S and LOCATE1.S are source files for Eric Isaacson's
;       excellent A86 assembler.  Assemble by typing:
;       A86 LOCATE.S LOCATE1.S
;
;       You may copy or distribute these files by whatever means.
;       If you wish to distribute a modified version, please include
;       my name and version number, as well as your name and a
;       summary of changes made.
;
;       No warranty whatsoever.  Hey, I'm giving you the source;
;       fix it yourself!


radix 16

doscall macro                          ; call to dos int_21 with a
mov ah,#1                              ; one-byte function in ah
int 21
#em

doscall2 macro                         ; call to dos int_21 with a
mov ax,#1                              ; two-byte function in ax
int 21
#em

dosprint macro                         ; call to dos string-print function
mov dx,offset #1
mov ah,09
int 21
#em

zprint macro                           ; call to program's asciiz-print
mov di,offset #1                       ; subroutine
call zprint1
#em

jmp begin                              ; skip over variables


tree_seg:                              ; segment for 64k tree buffer
dw 0000

tree_end:                              ; offset of second consecutive null
dw 0000

tree_max:                              ; largest value which tree_end attains
dw 0000                                ;    (curiosity value only)

env_seg:                               ; segment of environment block
dw 0000

env_temp:                              ; used by path parse routine
dw 0000

path_flag:                             ; nonzero = search only directories
db 00                                  ; in path

exe_flag:                              ; nonzero = list only files of type
db 00                                  ; .COM, .EXE, .BAT

no_remov_flag:                         ; nonzero = don't look at removable
db 00                                  ; drives

no_attr_flag:                          ; 00 = display attributes (default)
db 00                                  ; 40 = no attributes, 80 = all info

dirs_flag:                             ; 10 = display directories, 01 = don't
db 00                                  ; display directories, 00 = show both

hidden_flag:                           ; nonzero = display only hidden or
db 00                                  ; system files

verbose_flag:                          ; nonzero = debugging output
db 00

misc_flags:                            ; 80 = no drive letters
db 00                                  ; 40 = no dupe_drives check
                                       ; 20 = summary only
                                       ; 10 = peter-mode display
                                       ; 08 = peter, current dir printed?
                                       ; 02 = user specified a directory
                                       ; 01 = parsing from environment

lines_max:                             ; lines to scroll before pausing:
db 00                                  ; 00 = no pauses

lines_count:                           ; used internally for scroll pauses
db 00

wide_count:                            ; used for /w listing
db 00

entry_count:                           ; number of files/directories found
dw 0000
entry_count_hi:
dw 0000

files_count:                           ; number of files found
dw 0000
files_count_hi:
dw 0000

dirs_count:                            ; number of directories found
dw 0000
dirs_count_hi:
dw 0000

hfiles_count:                          ; hidden/system files found:
dw 0000
hfiles_count_hi:
dw 0000

hdirs_count:                           ; hidden/system directories found:
dw 0000
hdirs_count_hi:
dw 0000

file_totals:                           ; total size of all files found
dw 0000
file_totals_hi:
dw 0000

max_count:                             ; maximum matches to report
dw 0000

drv1:                                  ; first drive to search
db 'A'

drv2:                                  ; last drive to search
db 'Z'

ddrv_table:                            ; bitmapped table of duplicate drives
db 04 dup 00

fn_buf:                                ; user's filespec goes here:
db 51 dup 00                           ; will finally hold just the dir name

fs_buf:                                ; filespec stripped from dir spec
db 0d dup 00

working:                               ; working filespec buffer
db 51 dup 00

bat_buf:                               ; stores the batch prefix
db 43 dup 00

working_end:
dw 0000

append_hack:                           ; used to hack around append, if
dw 0000                                ; present:


begin:
mov bl,80                              ; CHECK FOR A VALID VERSION OF DOS:
doscall2 3306                          ; dos true-version-number function
cmp bl,05
jb dos_old                             ; less than 5 is old
cmp bl,09
jb dos_okay                            ; 5 through 9 okay, 10+ is bad
dos_old:                               ; msdos is pre-5:
doscall2 3000                          ; do old-style dos version check
cmp al,03
ja dos_okay                            ; dos 4 or better is okay
jb dos_bad                             ; dos 2 or worse is NOT okay
cmp ah,14
jae dos_okay                           ; dos 3.20 or better is okay

dos_bad:
call append_restore
dosprint msg_dos_bad                   ; print error message
doscall2 4c05                          ; terminate program
doscall 00                             ; terminate if above failed (dos 1.x)
dos_okay:

mov ax,0b700                           ; IS APPEND INSTALLED?
int 2f
cmp al,0ff
jne no_append                          ; no, skip ahead
mov ax,0b706                           ; yes, get append state
int 2f
test bx,0001                           ; is append enabled?
je no_append                           ; no, skip ahead
mov w [append_hack],bx                 ; yes, save append state
and bx,0fffe                           ; clear 'enabled' bit
mov ax,0b707                           ; set append state
int 2f
no_append:                             ; append is not a problem now

push cs                                ; CONFIGURE MEMORY:
pop es                                 ; get code segment into .es
mov bx,end_code_seg                    ; and new size in .bx
doscall 4a                             ; for dos resize memory-block function
jnc resize_ok                          ; if error,
call append_restore
dosprint msg_resize_error              ;    print error message
doscall2 4c04                          ;    and abort
resize_ok:
mov sp,(end_of_code + 200)             ; move stack down into remaining space
mov bx,1000
doscall 48                             ; request 64kb from dos
jnc alloc_ok                           ; if any error,
call append_restore
dosprint msg_alloc_error               ;    print error message
doscall2 4c04                          ;    and abort
alloc_ok:
mov w [tree_seg],ax                    ; save segment for tree buffer
mov es,ax
mov dx,0080                            ; set dta to 0080h
doscall 1a                             ; (this is dos default)
mov dx,offset new_int24                ; grab the dos critical-error int.
doscall2 2524                          ; to disable 'abort-retry-fail' msg.
cmp b [0019],01                        ; is output redirected?
jne not_pausing                        ; if so, skip ahead
mov b [lines_max],16                   ; output is to screen, so set up for
mov b [lines_count],16                 ;    automatic scroll pausing:
not_pausing:

mov bl,01                              ; drive a:
doscall2 440e                          ; ioctl get-drive-map function
jc no_drv_b_fix                        ; if any error, skip this routine
cmp al,00                              ; only one floppy present in system?
je no_drv_b_fix                        ; if more than one, don't do b: fix
or b [ddrv_table],02                   ; don't search drive b:
mov bl,01                              ; drive a:
doscall2 440f                          ; map floppy as drive a:
no_drv_b_fix:

call parse_env                         ; handle any locate= environment var.

mov di,offset fn_buf                   ; COPY FILENAME INTO BUFFER:
mov si,0081                            ; start at beginning of command tail
parse_fn:
lodsb                                  ; get a character
call force_uc                          ; convert lowercase to upper
call test_space                        ; is it a space?
je parse_fn_sp                         ; if so, go deal with it
cmp di,offset fn_buf
jne parse_fn_0                         ; if the first character
cmp al,'/'                             ; is a slash
je early_switch
cmp al,'-'                             ; or a dash,
jne parse_fn_0
early_switch:                          ; this is an 'early switch' :
mov w [fn_buf+0],'.' by '*'
mov w [fn_buf+2],00  by '*'            ; use default filespec of *.*
jmp parse_sw                           ; and skip ahead to the switch parser

parse_fn_0:
cmp al,'/'                             ; is it a slash?
jne parse_fn_1
mov al,'\'                             ; if so, convert to a backslash
parse_fn_1:
cmp al,'\'                             ; is it a slash or backslash?
jne parse_fn_2
or b [misc_flags],02                   ; if yes, note:  user specified dir.
parse_fn_2:
mov b [di],al                          ; stash character in buffer
inc di
cmp di,(fn_buf + 50)                   ; is the buffer full yet?
jb parse_fn                            ; nope, loop back for more
call append_restore
dosprint msg_buffer_full               ; yes, print error message
doscall2 4c02                          ;     and abort
parse_fn_sp:                           ; space or end-of-line found:
jc parse_fn_cr                         ; if it was a space
cmp b [fn_buf],00                      ;    and no filename found yet,
je parse_fn                            ;    ignore it and loop back
parse_fn_cr:                           ; end of line, or end of filename:
cmp b [fn_buf],00                      ; is there a filename in the buffer?
jne parse_fn_okay                      ; yes, continue
mov w [fn_buf+0],'.' by '*'
mov w [fn_buf+2],00  by '*'            ; no, use default filespec of *.*

parse_fn_okay:                         ; GOT A FILENAME IN THE BUFFER:
cmp al,0d                              ; at the end of the command line?
jne parse_sw                           ; no, parse any command-line switches
jmp parse_done                         ; yes, done parsing command line
parse_sw:                              ; look for switches:
lodsb                                  ; get a character
call test_switch                       ; return, space or switch char?
jnc parse_sw1
jmp parse_done                         ; if return, we are done parsing
parse_sw1:
je parse_sw                            ; space or switch, ignore it
call force_uc                          ; otherwise, force uppercase
mov di,0ffff
parse_sw2:
inc di                                 ; look at next char. in switch table
cmp b [switches+di],00                 ; is it a null?
je parse_syn                           ; if so, syntax error
cmp b [switches+di],al                 ; is it equal to the current char?
jne parse_sw2                          ; if not, keep looking
shl di,01
mov bx,w [switch_routines+di]          ; found it:  get the routine's address
jmp bx                                 ; and go there

parse_syn:
jmp syntax                             ; any other switch = syntax error

switch_t:
mov b [path_flag],40                   ; /t = set path flag
jmp skip_word                          ; and skip any nonswitch characters

switch_x:
mov b [exe_flag],40                    ; /x = set exe flag
jmp skip_word                          ; and skip any nonswitch characters

switch_r:
mov b [no_remov_flag],40               ; /r = set no_remov_flag
jmp skip_word                          ; and skip any nonswitch characters

switch_v:
mov b [verbose_flag],40                ; /v = set verbose flag
jmp skip_word                          ; and skip any nonswitch characters

switch_n:
mov b [no_attr_flag],40                ; /n = set no_attr_flag
and b [misc_flags],0cf                 ; turn off summary and peter modes
jmp skip_word                          ; and skip any nonswitch characters

switch_h:
mov b [hidden_flag],40                 ; /h = set hidden_flag
jmp skip_word                          ; and skip any nonswitch characters

switch_m:
or b [misc_flags],40                   ; /m = skip test for duplicate drives
jmp skip_word                          ; skip any nonswitch characters

switch_s:                              ; /s = summary info only
or b [misc_flags],20                   ; set for summary mode
and b [misc_flags],0ef                 ; disable peter mode
mov b [no_attr_flag],00
jmp skip_word

switch_d:
call next_char                         ; look at character after the D:
cmp al,'+'
je parse_sw_dp                         ; plus, set dirs_flag to 10
cmp al,'-'
je parse_sw_dm                         ; minus, set dirs_flag to 01
mov b [dirs_flag],00                   ; otherwise, set dirs_flag to 00
jmp skip_word
parse_sw_dp:                           ; /d+ :  dirs_flag = 10
mov b [dirs_flag],10                   ; display directories only
inc si
jmp skip_word
parse_sw_dm:                           ; /d- :  dirs_flag = 01
mov b [dirs_flag],01                   ; never display directories
inc si
jmp skip_word

switch_f:
mov w [max_count],0000                 ; /f = max count of files to find
call next_char                         ; look at character after 'f'
cmp al,':'
jne parse_sw_f1                        ; if it is not a colon, no problem
inc si                                 ; if it is, just skip over it
parse_sw_f1:
call next_char                         ; get next character:
cmp al,'-'                             ; is it a minus?
je parse_sw_fy                         ; if so, exit without zero check
cmp al,'0'
jb parse_sw_fx                         ; is it a digit?
cmp al,'9'                             ; if not, exit /f parser
ja parse_sw_fx
inc si                                 ; if it is, bump the pointer
sub al,'0'                             ; convert digit to value
mov ah,00                              ; make it 16-bit
push ax                                ; and save it
mov ax,000a
mul w [max_count]                      ; multiply working value by 10
cmp dx,0000                            ; did the multiply cause an overflow?
jne parse_sw_fz                        ; if so, abort with an error message
mov w [max_count],ax
pop ax                                 ; get value of last digit
add w [max_count],ax                   ; and add it in
jc parse_sw_fz                         ; if overflow, abort
jnc parse_sw_f1                        ; otherwise, loop back for next char.
parse_sw_fx:                           ; no more digits:
cmp w [max_count],0000                 ; do we have a number?
jne parse_sw_fy                        ; if so, exit happily
inc w [max_count]                      ; if not, assume /f1
parse_sw_fy:    
jmp skip_word                          ; done with /f parser
parse_sw_fz:                           ; /f number too damn big:
call append_restore
dosprint msg_fz                        ; complain
doscall2 4c10                          ; and exit

switch_i:                              ; /i = display additional info
mov b [no_attr_flag],80                ; note this switch
and b [misc_flags],0cf                 ; disable summary mode and peter mode
jmp skip_word                          ; and skip any nonswitch characters

switch_b:                              ; /b:"prefix"
test b [misc_flags],01
je switch_b0
jmp error_env_switch
switch_b0:
and b [misc_flags],0cf                 ; disable summary mode and peter mode
mov b [bat_buf],00                     ; empty the batch prefix buffer
mov bx,0000                            ; no characters in batch buffer
switch_b1:                             ; scan for first quote mark:
lodsb                                  ; get the next character
call test_switch
jc switch_b_nada                       ; if carriage return, abort
je switch_b_nada                       ; if switch, abort
cmp al,'"'                             ; found the first quote yet?
jne switch_b1                          ; if not, loop back
switch_b2:                             ; found the first quote:
lodsb                                  ; get the next character
cmp al,0d                              ; is it return?
je switch_b_bad                        ; if so, bad string, so abort
cmp al,'"'                             ; is it the second quote?
je switch_b3                           ; if so, deal with that
mov b [bat_buf+bx],al                  ; otherwise, stash the character
inc bx                                 ; and increment the buffer pointer
cmp bx,0041                            ; batch buffer overflow yet?
jne switch_b2                          ; if not, loop back for another char.
call append_restore
dosprint msg_bat_ovf                   ; batch buffer overflow:  complain
doscall2 4c10                          ; and abort
switch_b3:                             ; found the second quote:
cmp b [bat_buf],00                     ; something's in the buffer, right?
je switch_b_x                          ; if not, just leave the buffer empty
mov al,b [bat_buf-1+bx]                ; look at the last character in buffer
cmp al,20                              ; was the last character a space?
je switch_b4                           ; if so, don't add a final space
cmp al,09                              ; was the last character a tab?
je switch_b4                           ; if so, don't add a final space
mov b [bat_buf+bx],20                  ; otherwise ...
inc bx                                 ; add a final space
switch_b4:
mov w [bat_buf+bx],0001                ; always terminate with an 01,00
mov b [no_attr_flag],40                ; and set for a plain display
switch_b_x:
jmp skip_word

switch_b_nada:                         ; /b specified, with no prefix:
mov b [bat_buf],00                     ; empty the batch prefix buffer
dec si                                 ; undo that last lodsb
jmp skip_word                          ; and carry on with parse

switch_b_bad:
call append_restore
dosprint msg_bad_bat                   ; complain
doscall2 4c10                          ; and exit

switch_w:
mov b [wide_count],05                  ; /w = wide peter display

switch_p:                              ; /p = peter display
switch_a:                              ; /a = peter display
or b [misc_flags],10                   ; set for peter_mode
and b [misc_flags],5f                  ; no buck mode, no summary mode
mov b [bat_buf],00                     ; empty the batch buffer
jmp skip_word                          ; and skip any nonswitch characters

switch_buck:
or b [misc_flags],80                   ; set bit 7 in misc_flags
and b [misc_flags],0ef                 ; disable peter mode
mov b [no_attr_flag],40                ; don't display attributes
jmp skip_word                          ; and continue

next_char:
test b [misc_flags],01
jne next_char_e
mov al,b [si]
ret
next_char_e:
mov al,b [es:si]
ret

skip_word:                             ; valid switch:  skip remaining chars
test b [misc_flags],01
jne skip_env_word
lodsb                                  ; get a character
call test_switch                       ; return, switch, space?
jc parse_done                          ; return = done parsing
jne skip_word                          ; non-switch, non-space = lose it
jmp parse_sw                           ; otherwise, loop back for next switch
skip_env_word:
jmp parse_env_2

parse_done:                            ; DONE READING IN COMMAND LINE:

strip_drv:                             ; STRIP AWAY ANY DRIVE LETTER:
cmp b [fn_buf+1],':'                   ; is the second character a colon?
jne no_drv_spec                        ;    no, skip this section
mov al,b [fn_buf]                      ; get the character before the colon
cmp al,'A'
jb bad_drv_spec                        ; is it a letter?
cmp al,'Z'                             ; if not, exit with an error message
ja bad_drv_spec
mov b [drv1],al                        ; drive letter specified:
mov b [drv2],al                        ; save it as low and high drives
cmp b [fn_buf+2],00                    ; anything after the drive letter?
jne drv_spec_1
mov w [fn_buf+0],'.' by '*'            ; nothing after the drive letter:
mov w [fn_buf+2],00 by '*'             ; use default filespec of *.*
jmp no_drv_spec                        ; and skip ahead
drv_spec_1:
mov si,offset fn_buf + 2               ; something after the drive letter:
mov di,offset fn_buf                   ; move it down two bytes
drv_spec_1a:
mov al,b [si]
mov b [di],al
inc si
inc di
cmp al,00
jne drv_spec_1a
jmp no_drv_spec
bad_drv_spec:                          ; drive letter wasn't a letter:
call append_restore
dosprint msg_bad_drv                   ; print an error message
doscall2 4c02                          ; and abort
no_drv_spec:                           ; DONE STRIPPING AWAY DRIVE LETTER:

dots_fix:                              ; FIX FOR FILESPEC OF JUST '.' OR '..'
cmp b [fn_buf],'.'                     ; does the filespec start with '.' ?
jne dots_fix_x                         ; if not, never mind
cmp b [fn_buf+1],00
jne dots_fix_1                         ; filespec of just . ?
mov w [fn_buf+1],00 by '\'             ; if so, change to .\
jmp dots_fix_x
dots_fix_1:
cmp w [fn_buf+1],00 by '.'
jne dots_fix_x                         ; filespec of just .. ?
mov w [fn_buf+2],00 by '\'             ; if so, change to ..\
dots_fix_x:

mov si,offset fn_buf                   ; LOOK FOR DIRECTORY SPEC:
mov bl,'\'
call find_last                         ; find the last \ within filename buf.
jne dir_spec                           ; if found, skip ahead
mov si,offset fn_buf
mov di,offset fs_buf                   ; no directory was specified:
no_dir_spec_1:
mov al,b [si]                          ; copy filename
mov b [di],al                          ; into filespec buffer
inc si
inc di
cmp al,00
je no_dir_spec_2
cmp di,offset fs_buf + 0d              ; filespec longer than 12 chars?
jb no_dir_spec_1                       ; no, continue copying
call append_restore
dosprint msg_buffer_full               ; yes, print an error message
doscall2 4c02                          ; and abort
no_dir_spec_2:                         ; done copying filename:
mov w [fn_buf],00 by '\'               ; use default directory of '\'
jmp dir_spec_x                         ; and exit

dir_spec:                              ; A DIRECTORY WAS SPECIFIED:
mov dx,ax                              ; save offset of last '\'
mov si,ax
inc si                                 ; look at character after last '\':
cmp b [si],00                          ; is there any filespec?
jne dir_spec_1                         ; yes, skip ahead
mov w [fs_buf+0],'.' by '*'            ; directory specified, no filespec:
mov w [fs_buf+2],00 by '*'             ; use default of *.*
jmp dir_spec_x
dir_spec_1:                            ; directory and filespec both spec'ed:
mov di,offset fs_buf
dir_spec_2:
mov al,b [si]                          ; copy filename
mov b [di],al                          ; into filespec buffer
inc si
inc di
cmp al,00
je dir_spec_3
cmp di,offset fs_buf + 0d              ; filespec longer than 12 chars?
jb dir_spec_2                          ; no, continue copying
call append_restore
dosprint msg_buffer_full               ; yes, print an error message
doscall2 4c02                          ; and abort
dir_spec_3:                            ; done copying filename:
push dx
pop si
inc si                                 ; change character after last '\'
mov b [si],00                          ; to a null

dir_spec_x:                            ; CHECK DIR SPEC FOR WILDCARDS:
mov bl,'?'                             ; is there a question mark
mov si,offset fn_buf                   ;    in the directory name buffer?
call find_first
jne dir_spec_wc                        ; if so, error
mov bl,'*'                             ; is there an asterisk
mov si,offset fn_buf                   ;    in the directory name buffer?
call find_first
je dir_spec_xx                         ; if not, no wildcards, all is okay
dir_spec_wc:                           ; wildcard in directory name:
call append_restore
dosprint msg_dir_wc                    ;    print error message
doscall2 4c02                          ;    and abort

dir_spec_xx:                           ; DONE MESSING WITH DIRECTORY SPEC:
mov bl,'.'
mov si,offset fs_buf                   ; look for '.' in filespec
call find_first
jne dot_found_1
mov si,offset fs_buf                   ; there is no period:
call find_null                         ; find the terminal null
cmp si,(offset fs_buf + 8)             ; is it more than 8 characters after
jbe dot_not_found_1                    ;    the start of the filespec?
mov si,(offset fs_buf + 8)             ; if so, move it back
dot_not_found_1:
mov w [si],'*' by '.'                  ; dot was not found, add '.*'
inc si
inc si
mov b [si],00                          ; and terminal null to filespec
jmp dot_found_x
dot_found_1:                           ; there is a period:
cmp si,(offset fs_buf + 8)             ; is it more than 8 characters after
jbe dot_found_2                        ;    the start of the filespec?
mov di,(offset fs_buf + 8)
call copy_si_di_z                      ; if so, move the dot and ext. down
dot_found_2:
mov bl,'.'
mov si,offset fs_buf                   ; find the dot's position (again)
call find_first
add si,0004                            ; add 4
mov b [si],00                          ; poke in a null (ext. 3 chars max)
cmp b [exe_flag],00                    ; was /X specified?
je dot_found_x
sub si,0003                            ; if so, extension is always .*
mov w [si],00 by '*'                   ; so poke it in

dot_found_x:                           ; DONE MESSING WITH FILESPEC EXTENSION
mov si,offset fs_buf                   ; find the first asterisk in filespec
wc_fix:
mov bl,'*'
call find_first
je wc_fix_x                            ; no asterisk, skip ahead
inc si                                 ; look at first character after *
cmp b [si],00                          ; is it a null?
je wc_fix_x                            ;    yes, all is well
cmp b [si],'.'                         ; is it a period?
je wc_fix_p                            ;    yes, this is legal too
call append_restore
dosprint msg_wc_fix                    ; anything else is illegal --
doscall2 4c02                          ; print an error message and abort
wc_fix_p:                              ; asterisk before period:
inc si                                 ; move ahead to extension
jmp wc_fix                             ; and keep looking for asterisks
wc_fix_x:

                                       ; DONE MESSING WITH FILESPEC WILDCARDS
test b [misc_flags],02                 ; was a directory name specified?
je drv_fix_x                           ; if not, skip drive fix
mov al,b [drv1]
cmp al,b [drv2]                        ; was a drive specified?
je drv_fix_x                           ; if so, skip drive fix
doscall 19                             ; directory was spec'd, but no drive:
add al,'A'                             ; use current drive
mov b [drv1],al                        ;    for first drive
mov b [drv2],al                        ;    and last drive
drv_fix_x:

buck_fix:                              ; was /$ specified?
test b [misc_flags],80
je buck_fix_x                          ; if not, never mind
mov al,b [drv1]
cmp al,b [drv2]                        ; was a drive specified?
je buck_fix_x                          ; if so, never mind
doscall 19                             ; /$ specified, no particular drive:
add al,'A'                             ; use current drive
mov b [drv1],al                        ;    for first drive
mov b [drv2],al                        ;    and last drive
buck_fix_x:                            ; done with /$ fix

                                       ; DONE WITH POST-PARSE:
call dupe_drives                       ; check for duplicate drive mappings
cmp b [verbose_flag],00                ; in verbose mode?
je alls_well_1                         ; if not, skip ahead
dosprint msg_dir_is                    ; DISPLAY DEBUGGING INFO:
zprint fn_buf                          ; display directory spec
dosprint msg_fn_is
zprint fs_buf                          ; display file spec
dosprint msg_drv_is
mov dl,b [drv1]
doscall 02                             ; display first drive
dosprint msg_dash
mov dl,b [drv2]
doscall 02                             ; display second drive
dosprint msg_crlf
dosprint msg_max_is
mov ax,w [max_count]
mov ch,00
call dec_print_small                   ; display max count
dosprint msg_crlf
alls_well_1:                           ; done with debugging info:
cmp b [bat_buf],00                     ; is there a batch prefix?
je alls_well_2
zprint msg_bat_begin                   ; if so, print the '@echo off' message
alls_well_2:
test b [no_attr_flag],7f               ; in batch mode?
jne start_locate                       ; if not,
zprint msg_null                        ; output a blank line
cmp b [no_attr_flag],80                ; was /i specified?
jne start_locate
call zprint_add                        ; if so, output another

start_locate:                          ; ABOUT TO START TREE CONSTRUCTION:
cmp b [path_flag],00                   ; was /t specified?
je start_loc_0                         ; if not, continue
jmp use_path                           ; if so, we use an entirely different
start_loc_0:                           ;    routine to construct the tree
cmp b [no_remov_flag],00               ; was /r specified?
je start_loc_2                         ; if not, skip the following tests:
mov bl,b [drv1]                        ; get current drive
sub bl,40                              ; and convert to a 1-based number
doscall2 4408                          ; ioctl:  is drive removable?
jnc start_loc_1                        ; if call fails, skip this drive
start_loc_not:
jmp next_drive
start_loc_1:                           ; ioctl call succeeded:
test al,01                             ; is drive removable?
je start_loc_not                       ; if so, skip this drive
mov bl,b [drv1]                        ; get current drive
sub bl,40                              ; and convert to a 1-based number
doscall2 4409                          ; ioctl:  is drive remote?
jc start_loc_not                       ; if call fails, skip this drive
test dh,90                             ; is drive remote or subst'ed?
jne start_loc_not                      ; if so, skip this drive

start_loc_2:                           ; do use this drive:
mov di,0000
mov es,w [tree_seg]                    ; start of tree buffer:
mov al,b [drv1]                        ; poke in drive letter
mov ah,':'                             ; and a colon
mov b [msg_looking_1],al
mov w [es:di],ax
add di,0002
mov si,offset fn_buf                   ; then copy in user's directory spec
call copy_si_esdi_z                    ; (normally just a backslash, null)
mov b [es:di],00                       ; finish up with a second null
mov w [tree_end],di                    ; and save offset of 2nd null
cmp b [verbose_flag],00                ; being verbose?
je main_loop                           ; if not, skip drive message
zprint msg_looking                     ; if so, print drive message

main_loop:                             ; MAIN LOCATE LOOP:
mov di,offset working
mov es,w [tree_seg]
mov si,0000                            ; copy the first directory in tree
call copy_essi_di_z                    ; buffer to the working buffer
dec di
mov w [working_end],di                 ; save the offset of dirname's null
mov si,offset star_dot_star            ; add a filespec of *.*
call copy_si_di_z                      ; to the directory name
and b [misc_flags],0f7                 ; current directory not yet displayed
mov dx,offset working
mov cx,0016                            ; include hidden, system, subdirs
doscall 4e                             ; in dos find-first-match function
jnc found_something                    ; did dos find a match?
jmp next_dir                           ; no, use the next dir in tree buffer
found_something:                       ; yes, dos found a match:
cmp b [009e],'.'                       ; was it the stupid . or .. entry?
je next_file                           ; if so, ignore it
mov si,009e                            ; copy the filename
mov di,w [working_end]                 ; into the working buffer
call copy_si_di_z                      ; right after the directory name:
test b [0095],10                       ; is the item a subdirectory?
je add_tree_not                        ; if not, don't add it to tree buffer
cmp b [path_flag],00                   ; are we searching %path% only?
jne add_tree_not                       ; if so, don't add it to tree buffer
mov di,w [tree_end]                    ; ADD THIS DIRECTORY TO TREE BUFFER:
mov si,offset working                  ; copy the working buffer
call copy_si_esdi_z                    ; to the end of the tree buffer
dec di
mov w [es:di],00 by '\'                ; then add a backslash
add di,0002
mov b [es:di],00                       ; and a double-null
mov ax,di                              ;    new code to watch tree_end:
cmp ax,w [tree_end]                    ;    has pointer overflowed?
jae add_tree_20                        ;    if not, continue
jmp err_tree_ov                        ;    if so, barf and die
add_tree_20:
cmp ax,w [tree_max]                    ;    has tree_end hit a new high?
jbe add_tree_21
mov w [tree_max],ax                    ;    if so, save the new high
add_tree_21:
mov w [tree_end],di                    ; save offset of the second null
add_tree_not:
call test_filename                     ; print filename if it matches spec.
next_file:                             ; look for next file in this directory
cmp b [0019],01                        ;    was output redirected?
jne next_file1                         ;    if so, skip keyboard check
mov ah,01
int 16                                 ;    was a key pressed?
je next_file1                          ;    if not, do look for next file
mov b [lines_max],16                   ;    key was hit:  enable scroll-
mov b [lines_count],16                 ;    pausing (it might be off)
call pause                             ;    and pause for user keystrokes
next_file1:                            ; look for next file in this directory
doscall 4f                             ; dos find-next function
jc next_dir                            ; if no more matches, do next dir
jmp found_something                    ; if another match, loop back, do it
next_dir:
call peter_line_fix                    ; do line feeds as needed
mov si,0000                            ; start at beginning of tree buffer
call find_es_null                      ; and look for the first null
inc ax                                 ; is the character after it
cmp ax,w [tree_end]                    ; the final, tree-ending null?
je next_drive                          ; if so, skip ahead, do next drive
mov si,ax                              ; if not, more directories in tree buf
mov di,0000                            ; so copy entire tree buffer down
call copy_essi_esdi_zz                 ; to overwrite first directory in it
mov w [tree_end],di                    ; save new end-of-tree pointer
jmp main_loop                          ; and loop back to do new 1st dir.
next_drive:                            ; THE TREE BUFFER HAS BEEN EXHAUSTED:
cmp b [path_flag],00                   ; is this is /t path search?
jne final_exit                         ; if so, exit program now
mov al,b [drv1]
cmp al,b [drv2]                        ; is this the final drive?
je final_exit                          ; if so, exit program
inc al                                 ; increment drive letter
mov b [drv1],al                        ; save new drive letter
call drive_bit                         ; convert drive letter to bit pointer
test b [ddrv_table+bx],ah              ; test bit in table:  duplicate drive?
jne next_drive                         ; if so, skip it
jmp start_locate                       ; loop back, start a new tree

final_exit_1go:
jmp final_exit_1

final_exit:                            ; DONE SEARCHING FOR FILES:
mov ax,w [entry_count]
or ax,w [entry_count_hi]
cmp ax,0000                            ; were any files found?
jne final_1
jmp nothing_found                      ; if not, skip ahead ...
final_1:
test b [no_attr_flag],7f               ; allowed to print extra info?
jne final_exit_1go                     ; if not, skip ahead
call peter_line_fix
test b [misc_flags],10
jne final_1a
zprint msg_null                        ; some matches, print a blank line
final_1a:
mov dx,w [entry_count_hi]
mov ax,w [entry_count]                 ; get the big count value of entries
mov ch,00
call dec_print_big                     ; and print it out
mov w [msg_num_entries+6],ax
zprint msg_num_entries                 ; 'entries' message

mov dx,w [files_count_hi]
mov ax,w [files_count]                 ; get the big count value of files
mov ch,00
call dec_print_big                     ; and print it out
mov b [msg_num_files+5],ah
zprint msg_num_files                   ; 'files' message
cmp b [hidden_flag],00                 ; displaying only hidden or system?
jne zero_hfiles                        ; if so, skip over h/s count display
mov ax,w [hfiles_count]
or ax,w [hfiles_count_hi]              ; were any h/s files found?
cmp ax,0000
je zero_hfiles                         ; if not, skip over h/s count display
zprint msg_hs
mov dx,w [hfiles_count_hi]
mov ax,w [hfiles_count]
mov ch,00
call dec_print_big                     ; display number of h/s files found
zprint msg_hs1
zero_hfiles:                           ; done with h/s files count display
zprint msg_num_files1

mov dx,w [dirs_count_hi]
mov ax,w [dirs_count]                  ; get the big count of directories
mov ch,00
call dec_print_big                     ; and print it out
mov b [msg_num_dirs+0a],ax
cmp ax,0ffff
jne num_dirs_plural
mov b [msg_num_dirs+9],'y'
num_dirs_plural:
zprint msg_num_dirs                    ; 'directories' message
cmp b [hidden_flag],00                 ; displaying only hidden or system?
jne zero_hdirs                         ; if so, skip over h/s count display
mov ax,w [hdirs_count]
or ax,w [hdirs_count_hi]               ; were any h/s dir's found?
cmp ax,0000
je zero_hdirs                          ; if not, skip over h/s count display
zprint msg_hs
mov dx,w [hdirs_count_hi]
mov ax,w [hdirs_count]
mov ch,00
call dec_print_big                     ; display number of h/s dir's found
zprint msg_hs1
zero_hdirs:                            ; done with h/s dir's count display
zprint msg_num_dirs1

cmp w [max_count],0000                 ; was /f specified?
je final_exit_1a                       ; if not, we're done
mov ax,w [max_count]
cmp ax,w [entry_count]                 ; did we actually reach the /f max?
jne final_exit_1a                      ; if not, we're done
push ax
dosprint msg_maybe_more                ; if so, warn before finishing
pop ax
mov ch,00
call dec_print_small
zprint msg_maybe_more1
final_exit_1a:
mov ax,w [files_count]
or ax,w [files_count_hi]
cmp ax,0000
je final_exit_1
dosprint msg_file_totals
mov ax,w [file_totals]
mov w [009a],ax
mov ax,w [file_totals_hi]
mov w [009c],ax
call show_info_filesize
final_exit_1:
call show_tree_max
call append_restore
doscall2 4c00                          ; and exit with errorlevel 00
nothing_found:
test b [no_attr_flag],7f               ; allowed to print extra info?
jne final_exit_2                       ; if not, skip ahead
zprint msg_nothing                     ; no matches, print a sorry message
final_exit_2:
call show_tree_max
call append_restore
doscall2 4c01                          ; and exit with errorlevel 01

show_tree_max:                         ; DISPLAY MAXIMUM VALUE OF TREE_END:
cmp b [verbose_flag],00
je show_tree_max_x                     ; only in verbose mode
dosprint msg_tree_max                  ; print explanatory message
mov ax,w [tree_max]
mov ch,00
call dec_print_small                   ; and value of tree_max
zprint msg_null                        ; end-of-line
show_tree_max_x:
ret

test_filename:                         ; SHOULD WE PRINT OUT THIS FILENAME?
mov si,w [working_end]                 ; start of found file's name
mov di,offset fs_buf                   ; start of user's filespec
test_fn_1:                             ; loop to compare name with user spec.
mov al,b [si]                          ; al = byte from actual filename
mov ah,b [di]                          ; ah = byte from user filespec
cmp al,ah                              ; do they match?
je test_fn_m                           ; yes!  that's easy!  continue
cmp al,'.'                             ; period in filename?
je test_fn_p                           ; yes, go to special handler for that
cmp al,00                              ; end of filename?
je test_fn_z                           ; yes, go to special handler for that
cmp ah,'?'                             ; question mark in user filespec?
je test_fn_m                           ; yes, accept character
cmp ah,'*'                             ; asterisk in user filespec?
je test_fn_m2                          ; yes, accept character, don't bump di
jmp test_fn_n                          ; mismatch!  don't print this filename
test_fn_p:                             ; period handler:
cmp ah,'*'                             ; at an asterisk in user filespec?
jne test_fn_n                          ; if not, filename mismatch
inc di                                 ; yes, advance di, maybe find a period
jmp test_fn_1                          ; in the user's filespec ...?
test_fn_n:                             ; filename mismatch:
ret                                    ; do nothing, return to caller
test_fn_z:                             ; end of filename:
cmp w [di],00 by '.'                   ; end of no-extension filespec?
je test_fn_y                           ; yes, filename matches.  take it
cmp w [di],'*' by '.'                  ; end of any-extension filespec?
je test_fn_y                           ; yes, filename matches.  take it
cmp w [di],00 by '*'                   ; end of filespec with no extension?
je test_fn_y                           ; take it.  (should never happen?)
cmp w [di],'.' by '*'                  ; stuck at asterisk before period?
jne test_fn_n                          ; if not, this is a definite mismatch
add di,0002                            ; look at the extension:
cmp b [di],00                          ; is this a no-extension filespec?
je test_fn_y                           ; yes, filespec matches.  take it
cmp w [di],00 by '*'                   ; is this an any-extension filespec?
je test_fn_y                           ; yes, filespec matches.  take it
jmp test_fn_n                          ; otherwise, filename doesn't match
test_fn_m:                             ; this character matches:
inc di                                 ; increment pointers
test_fn_m2:                            ; (don't increment di if at an *)
inc si
cmp al,00                              ; hit the end of the filename / spec?
jne test_fn_1                          ; if not, loop back for next chars.

test_fn_y:                             ; FILENAME MATCHES USER FILESPEC:
cmp b [exe_flag],00                    ; was /x specified?
je show_attr                           ; if not, skip following /x tests
test b [0095],10                       ; is this file a directory?
jne test_fn_n                          ; if so, no match.
mov si,w [working_end]                 ; in the filename
mov bl,'.'                             ; search for the period
call find_first                        ; found a period?
je test_fn_n                           ; if not, reject this filename
mov si,ax
inc si                                 ; point just past the period:
cmp w [si],'O' by 'C'                  ; is the extension .COM ?
jne exe_chk_1                          ; if not, skip ahead, try .EXE
add si,0002
cmp b [si],'M'
jne test_fn_n
jmp show_attr                          ; extension is .COM, show it
exe_chk_1:
cmp w [si],'X' by 'E'                  ; is the extension .EXE ?
jne exe_chk_2                          ; if not, skip ahead, try .BAT
add si,0002
cmp b [si],'E'
jne test_fn_n
jmp show_attr                          ; extension is .EXE, show it
exe_chk_2:
cmp w [si],'A' by 'B'                  ; is the extension .BAT ?
jne test_fn_n                          ; if not, don't display this file
add si,0002
cmp b [si],'T'
jne test_fn_n

show_attr:                             ; FILENAME IS ACCEPTABLE:
cmp b [dirs_flag],00                   ; do we care about the directory bit?
je show_attr_1                         ; if not, skip following test
mov al,b [0095]                        ; get the file attribute byte
and al,10                              ; and mask off all but the subdir bit
mov ah,b [dirs_flag]                   ; get the directory-caring byte
and ah,10                              ; and mask off all but the subdir bit
cmp al,ah                              ; do they match?
je show_attr_1                         ; if so, display this file/dir
jmp attr_done_1                        ; if not, don't display it
show_attr_1:
cmp b [hidden_flag],00                 ; was /h specified?
je show_attr_2                         ; if not, skip following test
test b [0095],06                       ; is file either hidden or system?
jne show_attr_2
jmp attr_done_1                        ; if neither, don't display
show_attr_2:                           ; FILENAME AND ATTRIB ARE ACCEPTABLE:
test b [misc_flags],10                 ; are we in peter mode?
jne show_peter                         ;    if so, use peter print routine
test b [misc_flags],20                 ; supposed to display anything at all?
jne name_done_a                        ; if not, merely count this entry
cmp b [no_attr_flag],00                ; supposed to display attributes?
jne attr_done                          ; if not, skip ahead
call display_attributes
attr_done:                             ; ready to display filename:
cmp b [bat_buf],00                     ; is there a batch prefix?
je no_bat_buf
zprint bat_buf                         ; if so, print it first
no_bat_buf:
mov di,offset working                  ; point to the filename:
test b [misc_flags],80                 ; was /$ specified?
je not_buck
inc di                                 ; if so, skip over the drive letter
inc di
not_buck:
call zprint1                           ; display the filename
cmp b [no_attr_flag],80                ; /i specified?
jne name_done_a                        ; no, skip ahead
call show_info                         ; yes, show additional info
name_done_a:
jmp name_done

show_peter:                            ; PETER-STYLE FILENAME OUTPUT:
test b [misc_flags],08                 ; has the current dir been displayed?
jne peter_dir_done                     ; if so, don't display it again
mov bx,offset working                  ; point to start of directory name
mov cx,w [working_end]
cmp cx,(working+3)
je show_peter_1  
dec cx
show_peter_1:
mov ah,02
peter_dir:
mov dl,b [bx]                          ; get a char from directory name
int 21                                 ; and print it out
inc bx
cmp bx,cx                              ; hit the end of the directory name?
jb peter_dir                           ; if not, loop back
zprint msg_null
or b [misc_flags],08                   ; note that dir was printed
peter_dir_done:
cmp b [wide_count],00
je peter_long
mov dl,20
mov ah,02
int 21
int 21
mov cl,0d
mov bx,w [working_end]
peter_wide_1:
mov al,b [bx]
cmp al,00
je peter_wide_2
call peter_case
mov dl,al
doscall 02
inc bx
dec cl
jmp peter_wide_1
peter_wide_2:
mov ah,02
mov dl,20
peter_wide_3:
int 21
dec cl
cmp cl,00
jne peter_wide_3
dec b [wide_count]
cmp b [wide_count],00
jne peter_wide_4
mov b [wide_count],05
zprint msg_null
peter_wide_4:
jmp name_done
peter_long:
mov dl,20
mov cl,04
mov ah,02
peter_long_1:
int 21
dec cl
cmp cl,00
jne peter_long_1
mov bx,w [working_end]
mov cl,0f
peter_long_2:
mov al,b [bx]
cmp al,00
je peter_long_3
call peter_case
mov dl,al
int 21
inc bx
dec cl
jmp peter_long_2
peter_long_3:
mov dl,20
peter_long_4:
int 21
dec cl
cmp cl,00
jne peter_long_4
call show_info_00
jmp name_done

name_done:
inc w [entry_count]
jne attr_done_0
inc w [entry_count_hi]                 ; increment the big entry count
attr_done_0:
test b [0095],10                       ; file or directory?
jne attr_done_dir
inc w [files_count]
jne add_filesize
inc w [files_count_hi]                 ; increment the big file count
add_filesize:
mov ax,w [009a]
add w [file_totals],ax
mov ax,w [009c]
adc w [file_totals_hi],ax
test_hfiles:
test b [0095],06                       ; is this file hidden or system?
je attr_done_1                         ; if not, don't bump h/s files count
inc w [hfiles_count]
jne attr_done_1
inc w [hfiles_count_hi]                ; increment the big h/s files count
jmp attr_done_1
attr_done_dir:
inc w [dirs_count]
jne test_hdirs
inc w [dirs_count_hi]                  ; increment the big directory count
test_hdirs:
test b [0095],06                       ; is this directory hidden or system?
je attr_done_1                         ; if not, don't bump h/s dir's count
inc w [hdirs_count]
jne attr_done_1
inc w [hdirs_count_hi]                 ; increment the big h/s dir's count
attr_done_1:
cmp w [max_count],0000                 ; was there a /f specified?
je attr_done_2                         ; if not, exit
mov ax,w [max_count]                   ; /f was specified:
cmp ax,w [entry_count]                 ; have we shown enough entries?
jne attr_done_2                        ; if not, exit
pop ax                                 ; whoops, all done:  clear the stack
jmp final_exit                         ;   and i am outta here!
attr_done_2:
ret

display_attributes:
mov si,0000                            ; pointer to bit-name letters
mov cl,20                              ; highest bit is #5 (archive)
attr_loop:
mov al,'.'                             ; assume a space
test b [0095],cl                       ; is this bit set?
je attr_go                             ; clear, use the space
mov al,b [attr1+si]                    ; set, get the bit-name letter instead
attr_go:
inc si                                 ; point to the next letter
mov b [msg_attrib+si],al               ; and poke space or letter into string
attr_skip:
shr cl,1                               ; and shift right to next bit
cmp cl,08                              ; bit 3 (volume label/lfn) we don't
je attr_skip                           ; care about, so skip it
cmp cl,00                              ; run out of bits?
jne attr_loop                          ; no, loop back for next
dosprint msg_attrib                    ; yes, print the attribute string
ret

show_info:                             ; DISPLAY ADDITIONAL FILE INFO:
dosprint msg_info_1                    ; tab in a few spaces
show_info_00:
call display_attributes                ; display file's attributes
mov ax,w [0098]                        ; file's date stamp:
and ah,01
shr ax,05                              ; get month value in al
mov ah,80                              ; suppress leading zeroes
call time_print                        ; print month
dosprint msg_info_dash                 ; and a hyphen
mov ax,w [0098]                        ; file's date stamp:
and al,1f                              ; get date in al
mov ah,00                              ; print leading zeroes
call time_print                        ; print date
dosprint msg_info_dash                 ; and a hyphen
mov ax,w [0098]                        ; file's date stamp:
shr ax,09                              ; get year value in al
add ax,01980xd                         ; and normalize
mov ch,80
call dec_print_small                   ; print the year
dosprint msg_info_space                ; and tab in some
mov b [msg_info_ampm+1],'a'            ; assume a.m.
mov ax,w [0096]                        ; file's time stamp:
shr ax,0b                              ; get hour in al
cmp al,00                              ; is it midnight?
jne show_info_15                       ; no, continue ....
mov al,0c                              ; display midnight as '12 am'
jmp show_info_20
show_info_15:                          ; not midnight:
cmp al,0c                              ; is it before noon?
jb show_info_20                        ; if so, don't need to manipulate it
mov b [msg_info_ampm+1],'p'            ; noon or later:  display as 'pm'
cmp al,0c                              ; is it noon?
je show_info_20                        ; if so, display as '12 pm'
sub al,0c                              ; otherwise, normalize hours
show_info_20:                          ; done fiddling with hours:
mov ah,80                              ; suppress leading zeroes
call time_print                        ; and print hours
dosprint msg_info_colon                ; followed by a colon
mov ax,w [0096]                        ; file's time stamp:
and ah,07
shr ax,05                              ; get minutes in al
call time_print                        ; print minutes with leading zeroes
dosprint msg_info_ampm                 ; then 'am' or 'pm' indicator

test b [0095],10                       ; is this item a subdirectory?
jne show_info_30                       ; if so, branch ahead

show_info_filesize:
mov dx,w [009c]
mov ax,w [009a]                        ; file, not dir:  get the file's size
mov ch,00
call dec_print_big                     ; and display it
dosprint msg_info_bytes                ; as bytes
mov dx,w [009c]
mov ax,w [009a]
cmp dx,0010                            ; is the file 1 mb or larger?
jae show_info_mb                       ; if so, also show file size in mb
cmp dx,0000
jne show_info_kb                       ; is the file 1 kb or larger?
cmp ax,0400                            ; if so, also show file size in kb
jae show_info_kb
jmp show_info_32                       ; otherwise, byte size will suffice
show_info_30:
dosprint msg_info_dir                  ; directory, say so
jmp show_info_32                       ; and exit
show_info_kb:                          ; report file size in kb:
dosprint msg_info_open                 ; display the open parenthesis
mov dx,w [009c]
mov ax,w [009a]                        ; dword file size
shr ax,0a                              ; divide low word by 1024
shl dx,06                              ; get most significant bits
or ax,dx
mov ch,00
call dec_print_small                   ; and display kb value
mov ax,w [009a]                        ; get fractional kb
shr ax,02                              ; in al (0 to 255)
call show_fraction                     ; and display fractional kb
dosprint msg_info_kb                   ; close parenthesis
jmp show_info_32                       ; done
show_info_mb:                          ; report file size in mb:
dosprint msg_info_open                 ; display open parenthesis
mov ax,w [009c]
shr ax,04                              ; file size in mb
mov ch,00
call dec_print_small                   ; display mb value
mov ax,w [009a]
mov dx,w [009c]
shr ax,0c
shl dx,04
or ax,dx                               ; fractional mb value (0-255) in al
call show_fraction                     ; display fractional mb
dosprint msg_info_mb                   ; close parenthesis
show_info_32:
zprint msg_null                        ; done with file info, complete line
ret

show_fraction:                         ; al contains a value 0-255:
push ax                                ; save it
dosprint msg_info_dot                  ; display a decimal point
pop ax                                 ; al contains 0-255:
mov ah,00                              ; convert to word length
mov bl,64
mul bl                                 ; multiply by 100
mov al,ah                              ; then divide by 256
mov ah,00                              ; and fall through

time_print:
mov dx,'0' by '0'
time_print_1:
cmp al,0a
jb time_print_2
sub al,0a
inc dl
jmp time_print_1
time_print_2:
add dh,al
test ah,80
je time_print_3
cmp dl,'0'
jne time_print_3
mov dl,20
time_print_3:
mov ah,02
int 21
mov dl,dh
int 21
ret

peter_line_fix:
test b [misc_flags],08
je peter_line_x
zprint msg_null
and b [misc_flags],0f7
cmp b [wide_count],00
je peter_line_x
cmp b [wide_count],05
je peter_line_x
zprint msg_null
mov b [wide_count],05
peter_line_x:
ret

path_bad:                              ; no path variable:
jmp main_loop                          ; exit.  search only current directory

use_path:                              ; FILL TREE BUFFER FROM PATH VARIABLE:
doscall 19                             ; get the current drive
add al,'A'                             ; and convert to a letter
mov ah,':'                             ; add a colon
mov es,w [tree_seg]                    ; segment of tree buffer
mov si,0000                            ; at the beginning:
mov w [es:si],ax                       ; poke the drive letter and colon
add si,0002
mov b [es:si],'\'                      ; add a backslash
inc si
doscall 19                             ; get the current drive again
mov dl,al
inc dl                                 ; convert to a 1-based number in dl
mov ds,w [tree_seg]                    ; ds:si points right after the \
doscall 47                             ; add the current directory after \
push cs
pop ds                                 ; put ds back where god intended
mov es,w [tree_seg]                    ; es points to the tree buffer 
mov si,0000
call find_es_null                      ; find the terminal null
cmp ax,0004                            ; is the current directory the root?
jb use_path_fix
mov b [es:si],'\'                      ; if not, add a final backslash
inc si
use_path_fix:
mov w [es:si],0000                     ; add a double-null
inc si
mov w [tree_end],si                    ; and save offset of the 2nd null

mov ax,w [002c]                        ; program's psp contains
mov w [env_seg],ax                     ; segment of an environment block
cmp ax,0008                            ; is it above 8?
jna path_bad                           ; if not, something is very wrong

mov bp,offset var_path                 ; look for the path= variable
call find_env_var
jne path_bad                           ; if no path= variable, abort
mov w [env_temp],si                    ; save pointer

path_loop:
mov es,w [env_seg]
mov si,w [env_temp]                    ; pointer into path value
mov di,offset working                  ; pointer into working buffer
mov dl,00                              ; no last character
path_loop_1:
mov al,b [es:si]                       ; get a byte from path value
inc si                                 ; and increment pointer
call force_uc                          ; no lowercase, please
cmp al,20                              ; is it a space?
je path_loop_1                         ; if so, ignore it
cmp al,09                              ; is it a tab?
je path_loop_1                         ; if so, ignore it
cmp al,'/'                             ; is it a slash? (should never happen)
jne path_loop_2
mov al,'\'                             ; if so, substitute a backslash
path_loop_2:
cmp al,';'                             ; semicolon = end of directory name
je path_loop_3
cmp al,00                              ; null = end of directory name
je path_loop_3
mov b [di],al                          ; otherwise, copy into working buffer
inc di
mov dl,al                              ; and save this character as last char
jmp path_loop_1                        ; and loop back for next
path_loop_3:                           ; directory name in working buffer:
cmp di,offset working                  ; is there really?
je path_whoops                         ;    maybe not ...
cmp dl,'\'                             ; was the last character a backslash?
je path_loop_4
mov b [di],'\'                         ; if not, add a backslash to the end
inc di
path_loop_4:
mov w [env_temp],si                    ; save pointer into environment
mov b [di],00                          ; null-terminate the working dir name
mov dl,al                              ; and stash the semicolon or null
mov si,offset working                  ; compare the directory name
mov es,w [tree_seg]                    ; with the first directory in the
mov di,0000                            ; tree buffer (the current directory)
call compare_si_esdi_z                 ; are they identical?
je path_loop_5                         ; if so, don't add this directory
mov si,offset working                  ; probably not a duplicate:
mov di,w [tree_end]
call copy_si_esdi_z                    ; copy directory onto end of tree
mov b [es:di],00                       ; and add a second null
mov w [tree_end],di                    ; save address of second null
path_loop_5:
cmp dl,00                              ; found the null at end of path var?
jne path_loop                          ; no, loop back for next dir name
path_bad_2:
jmp main_loop                          ; done filling tree_buffer from path
path_whoops:                           ; working buffer is empty!  this may
cmp al,00                              ; happen with pathological paths.
je path_bad_2                          ; abort if end of variable reached;
mov w [env_temp],si                    ; otherwise, attempt to continue
jmp path_loop                          ; parsing path variable.

