
dupe_drives:                           ; CHECK FOR DUPLICATE DRIVE MAPPINGS:
test b [misc_flags],40                 ; did the user specify /M ?
jne ddrv1                              ; if so, exit this routine
cmp b [path_flag],00                   ; did the user specify /T ?
jne ddrv1                              ; if so, exit this routine
mov al,b [drv1]
cmp al,b [drv2]                        ; did the user specify a single drive?
jne ddrv2                              ; if not, continue
ddrv1:
ret
ddrv2:                                 ; initialize dupe-drive workspace:
mov ax,':' by 'a'                      ; start with drive a:
mov bx,0000                            ; and slot #0
mov es,w [tree_seg]                    ; use tree buffer for workspace
ddrv3:
mov w [es:bx],ax                       ; poke drive letter into slot
mov b [es:bx+2],00                     ; and zero-terminate it
add bh,01                              ; next slot
add al,01                              ; and next drive letter
cmp al,'z'                             ; run out of drive letters yet?
jna ddrv3                              ; if not, loop back
mov w [tree_end],0000                  ; initialize assign-list index
ddrv4:
mov bx,w [tree_end]                    ; get assign-list index
mov si,4000                            ; offset of local-name buffer
mov di,4100                            ; offset of network-name buffer
mov es,w [tree_seg]                    ; segment of network-name buffer
push es
pop ds                                 ; segment of local-name buffer
mov b [ds:si],00                       ; empty the local-name buffer
mov b [es:di],00                       ; empty the network-name buffer
doscall2 5f02                          ; get assign-list entry:
push cs
pop ds                                 ; get my data segment back!
jc ddrv11                              ; if out of entries, exit
mov es,w [tree_seg]                    ; segment of dupe-drive workspace
mov ax,w [es:4000]                     ; check the local name:
cmp ah,':'                             ; is the second character a colon?
jne ddrv10                             ; if not, continue with next entry
cmp al,'A'
jb ddrv10                              ; is the first character an uppercase
cmp al,'Z'                             ; letter?
ja ddrv10                              ; if not, continue with next entry
sub al,'A'                             ; convert drive letter to number
mov bh,al
mov bl,00                              ; bx contains 256*drive
mov si,4100                            ; si points to network-name buffer
mov cl,00                              ; no backslashes copied yet
ddrv6:
mov al,b [es:si]                       ; copy the network name
mov b [es:bx],al                       ; to the drive's slot (256*drive)
inc si
inc bx
cmp al,'\'                             ; was that a backslash?
jne ddrv7
inc cl                                 ; if so, count it
cmp cl,04                              ; was it the fourth backslash?
je ddrv8                               ; if so, force end of netname
ddrv7:
cmp al,00                              ; was character the terminal null?
jne ddrv6                              ; if not, loop back for next char
je ddrv10                              ; if so, done copying netname
ddrv8:                                 ; four backslashes:
mov b [es:bx],00                       ; null-terminate the destination name
ddrv10:                                ; increment assign-list entry
inc w [tree_end]
jmp ddrv4
ddrv11:                                ; DONE BUILDING DUPE-DRIVE WORKSPACE:
mov es,w [tree_seg]
mov w [tree_end],0100                  ; start upper drive at 1 (b:)
ddrv12:
mov w [env_temp],0000                  ; start lower drive at 0 (a:)
ddrv13:
mov bx,w [tree_end]                    ; bx points to upper drive slot
mov bp,w [env_temp]                    ; bp points to lower drive slot
mov si,0000                            ; start at beginning of netname
ddrv14:
mov al,b [es:bx+si]                    ; get a character from upper netname
cmp al,b [es:bp+si]                    ; compare against lower netname
jne ddrv16                             ; if mismatch, try next lower drive
inc si                                 ; character matches:
cmp al,00                              ; was it the terminal null?
jne ddrv14                             ; if not, loop back for next char.
mov al,b [tree_end+1]                  ; get the drive number
call drive_bit1                        ; and convert to a bit pointer
or b [ddrv_table+bx],ah                ; set the bit in the dupe drive table
jmp ddrv18                             ; and move on to the next upper drive
ddrv16:                                ; increment lower drive:
mov bx,w [env_temp]
add bx,0100
mov w [env_temp],bx
cmp bx,w [tree_end]
jne ddrv13
ddrv18:                                ; increment upper drive:
mov bx,w [tree_end]
add bx,0100
mov w [tree_end],bx
cmp bx,1a00
jb ddrv12
ret                                    ; done checking for dupe drives

drive_bit:                             ; CONVERT DRIVE LETTER TO BIT POINTER:
sub al,'A'                             ; al holds 0-based drive number
drive_bit1:
mov bl,al 
shr bl,03
mov bh,00                              ; bx holds drive/8
mov cl,al
and cl,07                              ; cl holds (drive and 07)
mov ah,01
shl ah,cl                              ; ah holds 2^(drive and 07)
ret                                    ; return byte ptr. in bx and bit in ah

parse_env:
mov ax,w [002c]                        ; program's psp contains
mov w [env_seg],ax                     ; segment of environment block
cmp ax,0008                            ; is it above 8?
ja parse_env_1                         ; if so, continue
parse_env_0:
ret                                    ; bad environment segment, abort
parse_env_1:
mov bp,offset var_locate
call find_env_var                      ; find any 'locate' variable
jne parse_env_0                        ; exit if not found
or b [misc_flags],01                   ; note: parsing from environment
parse_env_2:
mov al,b [es:si]                       ; get a character from environment:
call force_uc                          ; uppercase it
inc si
cmp al,00                              ; end of the environment var?
je parse_env_x                         ; if so, exit
call test_switch                       ; switch or space?
je parse_env_2                         ; if so, ignore it
mov di,0ffff
parse_env_4:
inc di                                 ; look at next char. in switch table
cmp b [switches+di],00                 ; is it a null?
je parse_env_syn                       ; if so, syntax error
cmp b [switches+di],al                 ; is it equal to the current char?
jne parse_env_4                        ; if not, keep looking
shl di,01
mov bx,w [switch_routines+di]          ; found it:  get the routine's address
jmp bx                                 ; and go there
parse_env_x:                           ; done parsing from enviroment block:
and b [misc_flags],0fe                 ; note it and exit
ret
parse_env_syn:                         ; illegal command in environment:
mov b [msg_env_err1],al
call append_restore
dosprint msg_syntax
dosprint msg_env_err
doscall2 4c10                          ; and abort

find_env_var:
cmp w [env_seg],0008                   ; valid segment for environment block?
ja find_env_var0                       ; if so, continue
mov al,00
cmp al,0ff                             ; if not,
ret                                    ; return with z clear for failure
find_env_var0:
mov es,w [env_seg]                     ; segment of environment block
mov bx,0000                            ; start of table:
find_env_var1:
mov si,0000
find_env_var2:
mov al,b [es:bx+si]                    ; get a character from env table
cmp al,b [bp+si]                       ; and compare against desired name
jne find_env_var8                      ; mismatch, find next var name
inc si
cmp al,'='                             ; found the terminal = ?
jne find_env_var2                      ; no, continue comparing names
mov ax,si
add bx,ax
mov si,bx                              ; return offset of char. after =
cmp al,al                              ; with z set for success
ret
find_env_var8:                         ; look for next variable name:
inc bx                                 ; scan forward
cmp b [es:bx],00                       ; until the first null
jne find_env_var8
inc bx                                 ; the character after the null:
cmp b [es:bx],00                       ; is it another null?
jne find_env_var1                      ; no, found the next var name -- cont.
cmp b [es:bx],0ff                      ; out of variables - exit with z clear
ret

error_env_switch:
dec si
mov al,b [es:si]
call force_uc
mov b [msg_env_err1],al
call append_restore
dosprint msg_env_err
doscall2 4c10

test_switch:                           ; EXAMINE CHARACTER IN .AL:
cmp al,'/'                             ; if a switch, exit with carry clear
je test_sw_1                           ;    and zero set
cmp al,'-'                             ; otherwise, return flags as in
je test_sw_1                           ;    following routine
cmp al,','
je test_sw_1

test_space:                            ; EXAMINE CHARACTER IN .AL:
cmp al,0d                              ; if return, exit with carry set
je test_sw_cr                          ; if space, exit with carry clear
cmp al,20                              ;    and zero set
je test_sw_1                           ; otherwise, exit with carry and
cmp al,09                              ;    zero clear
test_sw_1:
clc
ret
test_sw_cr:
stc
ret

peter_case:
test b [0095],10
je force_lc

force_uc:                              ; FORCE CHARACTER IN .AL TO UPPERCASE:
cmp al,'a'
jb force_uc_1
cmp al,'z'
ja force_uc_1
and al,0df
force_uc_1:
ret

force_lc:                              ; FORCE CHARACTER IN .AL TO LOWERCASE:
cmp al,'A'
jb force_lc_1
cmp al,'Z'
ja force_lc_1
or al,20
force_lc_1:
ret

find_last:                             ; SEARCH ASCIIZ STRING AT [SI] FOR
mov ax,0000                            ; LAST OCURRENCE OF CHARACTER IN BL:
find_last_1:
mov bh,b [si]                          ; if found, exit with offset of last
cmp bh,00                              ; match in ax and zf clear
je find_last_x
cmp bh,bl                              ; if not found, exit with ax=0000
jne find_last_2                        ; and zf set
mov ax,si
find_last_2:
inc si
jmp find_last_1
find_last_x:
cmp ax,0000
ret

find_first:                            ; SEARCH ASCIIZ STRING AT [SI] FOR
mov ax,0000                            ; FIRST OCURRENCE OF CHARACTER IN BL:
find_first_1:
mov bh,b [si]                          ; if found, exit with offset of first
cmp bh,00                              ; match in ax and si, and zf clear
je find_first_x
cmp bh,bl                              ; if not found, exit with ax=0000
jne find_first_2                       ; and zf set
mov ax,si
jmp find_first_x
find_first_2:
inc si
jmp find_first_1
find_first_x:
cmp ax,0000
ret

find_null:                             ; FIND FIRST NULL BYTE AFTER [SI]:
cmp b [si],00
je find_null_x                         ; return offset in si and ax
inc si
jmp find_null
find_null_x:
mov ax,si
ret

find_es_null:                          ; FIND FIRST NULL BYTE AFTER [ES:SI]:
cmp b [es:si],00
je find_null_x                         ; return offset in si and ax
inc si
jmp find_es_null

copy_si_di_z:                          ; COPY ASCIIZ STRING FROM [SI] TO [DI]
mov al,b [si]
mov b [di],al                          ; will malf if copying up to an
inc si                                 ; overlapping block
inc di
cmp al,00
jne copy_si_di_z
ret

copy_si_esdi_z:                        ; COPY ASCIIZ FROM [SI] TO [ES:DI]:
mov al,b [si]
mov b [es:di],al
inc si
inc di
cmp al,00
jne copy_si_esdi_z
ret

copy_essi_di_z:                        ; COPY ASCIIZ FROM [ES:SI] TO [DI]:
mov al,b [es:si]
mov b [di],al
inc si
inc di
cmp al,00
jne copy_essi_di_z
ret

copy_essi_esdi_zz:                     ; COPY BLOCK FROM [ES:SI] TO [ES:DI]:
mov al,b [es:si]                       ; block ends in null word (not byte)
mov b [es:di],al
inc si
inc di
cmp al,00
jne copy_essi_esdi_zz
cmp b [es:si],00
jne copy_essi_esdi_zz
mov b [es:di],00
ret

compare_si_esdi_z:                     ; COMPARE ASCIIZ STRING AT [SI] TO
mov al,b [si]                          ; STRING AT [ES:DI]:
cmp al,b [es:di]
jne compare_x                          ; returns zf set if strings are equal,
cmp al,00                              ; zf clear if strings are not equal
je compare_x
inc si
inc di
jmp compare_si_esdi_z
compare_x:
ret

lowercase_si_z:                        ; FORCE ASCIIZ STRING AT [SI] TO
mov al,b [si]                          ; LOWERCASE:
cmp al,00
je lowercase_x
cmp al,'A'
jb lowercase_1
cmp al,'Z'
ja lowercase_1
or al,20
mov b [si],al
lowercase_1:
inc si
jmp lowercase_si_z
lowercase_x:
ret

zprint1:                               ; PRINT ASCIIZ STRING, PAUSE IF NEEDED
mov ah,02                              ; dos print-character function
mov cl,00                              ; no ctrl-a character found
zprint2:
mov dl,b [di]                          ; get a character
inc di                                 ; and increment pointer
cmp dl,01                              ; found a control-a?
jne zprint3                            ; no, continue
mov cl,01                              ; yes, note it
jmp zprint2                            ; and loop back
zprint3:
cmp dl,00                              ; found the terminal null?
je zprintx                             ; if so, exit the print loop
cmp dl,0ff                             ; print-nothing character?
je zprint2                             ; if so, ignore it and loop back
int 21                                 ; if not, print the character
jmp zprint2                            ; and loop back for more
zprintx:                               ; found the terminal null:
cmp cl,01                              ; was there a control-a in the string?
je zprint_3                            ; if so, exit with no cr/lf
mov dl,0d
int 21                                 ; print a return
mov dl,0a
int 21                                 ; and a line feed
cmp b [0019],01                        ; is output redirected?
jne zprint_add                         ; if so, skip check for control key
mov ah,02
int 16                                 ; get keyboard shift status
test al,04                             ; was control pressed?
je zprint_add                          ; if not, don't do control-key delay
push es
mov es,0040                            ; segment for bios data area
keydelay1:
test b [es:006c],01                    ; wait for an even jiffy
jne keydelay1
keydelay2:
test b [es:006c],01                    ; then wait for an odd jiffy
je keydelay2
pop es
zprint_add:
cmp b [lines_max],00                   ; are we counting lines?
jne zprint_c                           ; yes, count this one
zprint_3:                              ; no, exit this routine
ret
zprint_c:                              ; counting lines:
dec b [lines_count]                    ; decrement line count
cmp b [lines_count],00                 ; time for a pause yet?
jne zprint_3                           ; no, exit routine
mov al,b [lines_max]                   ; time for a pause --
mov b [lines_count],al                 ; reset line count and fall through

pause:                                 ; PAUSE TO PREVENT SCROLLING:
dosprint msg_pause                     ; print hit-a-key message
pause_1:
mov ah,01
int 16                                 ; any keys in keyboard buffer?
je pause_2                             ; no, wait for one
mov ah,00                              ; yes, remove a key from buffer
int 16
jmp pause_1                            ; and loop back
pause_2:                               ; keyboard buffer is empty now:
mov ah,00
int 16                                 ; wait for a keystroke
call force_uc
cmp al,0d                              ; enter key?
je pause_cr                            ; yes, scroll line
cmp al,20                              ; space bar?
je pause_sp                            ; yes, scroll page
cmp ax,3d00                            ; f3 key?
je pause_brk                           ; yes, abort program now
cmp ax,0000                            ; break key?
je pause_brk                           ; yes, abort program now
cmp al,03                              ; control-c?
je pause_brk                           ; yes, abort program now
cmp al,1b                              ; escape?
je pause_brk                           ; yes, abort program now
cmp al,'C'                             ; c key?
je pause_esc                           ; yes, no more pauses
jne pause_2                            ; no, not a legal key, loop back

pause_esc:
mov b [lines_max],00                   ; escape pressed:  disable pausing
pause_sp:
dosprint msg_unpause                   ; erase the hit-a-key message
ret                                    ; and exit

pause_cr:
mov b [lines_count],01
jmp pause_sp

pause_brk:                             ; control-c hit:
call append_restore
dosprint msg_unpause
dosprint msg_abort
doscall2 4c03                          ; squiff program with errorlevel 03

dec_print_small:                       ; DISPLAY WORD IN AX:
mov dx,0000                            ; fall through ....

dec_print_big:                         ; DISPLAY DOUBLEWORD IN DX:AX
mov w [dp_t1],ax                       ; save the doubleword value
mov w [dp_t2],dx                       ; for later use
mov w [dp_t3],'s' by 'e'
cmp dx,0000
jne dec_print_plural
cmp ax,0001
jne dec_print_plural
mov w [dp_t3],0ffff
dec_print_plural:
mov bx,000a                            ; ten decimal is our divisor
push bx                                ; also our end-of-stack marker
mov cl,00                              ; init digits-between-commas count
dec_print_1:                           ; main decimal division loop
mov dx,0000
mov ax,w [dp_t2]                       ; dx:ax contains the high word as quad
div bx                                 ; divide by ten
mov w [dp_t2],ax                       ; save new high-order word
mov ax,w [dp_t1]                       ; get the old low word
div bx                                 ; and divide by ten
mov w [dp_t1],ax                       ; save new low word
add dx,0030                            ; convert remainder to ascii digit
push dx                                ; and save it on the stack
cmp ax,0000                            ; anything left?
jne dec_print_12                       ; if so, continue with divisions
cmp w [dp_t2],0000
je dec_print_19                        ; if not, print out digits
dec_print_12:
test ch,80                             ; supposed to print commas?
jne dec_print_1                        ; if not, loop back for next digit
inc cl                                 ; increment digits-between-commas
cmp cl,03                              ; three digits yet?
jne dec_print_1                        ; if not, loop back for next digit
mov cl,00                              ; yes, zero counter
mov dx,',' by ','                      ; and push a comma on the stack
push dx
jmp dec_print_1                        ; and loop back for next digit
dec_print_19:                          ; done with divisions:
mov ah,02                              ; dos print-character function:
dec_print_2:                           ; decimal printout loop
pop dx                                 ; get a digit from the stack
cmp dx,000a                            ; hit the end of stack yet?
je dec_print_3                         ; if so, exit decimal print routine
int 21                                 ; otherwise, print the digit
jmp dec_print_2                        ; and loop back for next digit
dec_print_3:
mov ax,w [dp_t3]
ret

dp_t1:                                 ; holds low-order word
dw 0000
dp_t2:                                 ; holds high-order word
dw 0000
dp_t3:                                 ; normally holds ascii 'es'
dw 0000

append_restore:                        ; RESTORE APPEND STATE, IF CHANGED:
cmp w [append_hack],0000               ; was append state changed?
je append_rest_x                       ; no, exit
mov bx,w [append_hack]                 ; yes, get original state
mov ax,0b707
int 2f                                 ; set append state
mov w [append_hack],0000               ; clear flag (for no good reason)
append_rest_x:

error_output:                          ; PREPARE TO PRINT AN ERROR MESSAGE:
mov bx,0002                            ; copy standard error handle
mov cx,0001                            ; to standard output handle
doscall 46                             ; dos force-duplicate-handle function
ret


err_tree_ov:                           ; TREE BUFFER OVERFLOW:
call append_restore
dosprint msg_tree_ov                   ; print an error message
doscall2 4c06                          ; and abort with errorlevel 06

syntax:                                ; SYNTAX ERROR:
call append_restore
dosprint msg_syntax                    ; print syntax message
doscall2 4c10                          ; and abort

new_int24:                             ; NEW CRITICAL-ERROR HANDLER:
mov al,03                              ; always fail
iret                                   ; pretty simple, huh?


switches:
db 'TXRVNHDFIBMSWPA$',00

switch_routines:
dw switch_t, switch_x, switch_r, switch_v, switch_n, switch_h
dw switch_d, switch_f, switch_i, switch_b, switch_m, switch_s
dw switch_w, switch_p, switch_a, switch_buck

msg_syntax:
db 0d,0a
db 'LOCATE.COM   v1.12   C. Dye   02-16-1996',0d,0a,0a
db 'Syntax:  LOCATE [filespec] [switches]',0d,0a,0a
db '    /H   list only hidden or system',0d,0a
db '    /X   list only files of type .COM, .EXE, .BAT',0d,0a
db '    /D-  list files only, not directories',0d,0a
db '    /D+  list directories only, not files (default is both)',0d,0a
db '    /T   search only current directory and path',0d,0a
db '    /R   no search on removable or network drives',0d,0a
db '    /Fn  find only first n matches',0d,0a
db '    /N   no display of attributes or counts',0d,0a
db '    /S   display summary info only, not names',0d,0a
db '    /W   Wide display',0d,0a
db '    /P or /A      Peter display',0d,0a
db '    /B:"command"  prefix command to filenames',0d,0a
db 0a
db 'Attributes (shown by default):',0d,0a
db '    A  Archive needed    S  System file    R  Read-only file',0d,0a
db '    D  subDirectory      H  Hidden file',0d,0a
db '$'

msg_resize_error:
db 0d,0a,'Error resizing program block!',0d,0a,'$'

msg_alloc_error:
db 0d,0a,'Error allocating tree buffer!',0d,0a,'$'

msg_dos_bad:
db 0d,0a,'This program requires DOS 3.2 or higher!',0d,0a,'$'

msg_buffer_full:
db 0d,0a,'Error:  Filespec too long!',0d,0a,'$'

msg_bad_drv:
db 0d,0a,'Illegal drive spec!',0d,0a,'$'

msg_dir_wc:
db 0d,0a,'Error:  Wildcards not allowed in directory name!',0d,0a,'$'

msg_wc_fix:
db 0d,0a,'Error:  Illegal use of * wildcard in filespec!',0d,0a,'$'

msg_fz:
db 0d,0a,'Error:  /Fn  -- number cannot be greater than 65535.',0d,0a,'$'

msg_bad_bat:
db 0d,0a,'Error:  Bad string after /B -- should be /B:"command"',0d,0a,'$'

msg_bat_ovf:
db 0d,0a,'Error:  Bad string after /B -- 64 characters max!',0d,0a,'$'

msg_tree_ov:
db 0d,0a,'Error:  Tree buffer overflow!  (Directory structure too complex)',0d,0a,'$'

msg_env_err:
db 0d,0a,'Error:  Illegal switch "'
msg_env_err1:
db '#" in LOCATE variable!',0d,0a,'$'

msg_dir_is:
db 0d,0a,'Directory:  $'

msg_fn_is:
db 'Filespec :  $'

msg_drv_is:
db 'Drives   :  $'

msg_max_is:
db 'Max_count:  $'

msg_dash:
db '-$'

msg_bat_begin:
db '@echo off',00

msg_crlf:
db 0d,0a,'$'

msg_pause:
db 0d,'  <Space> scroll page, <Enter> scroll line, <F3> abort, <C> continous :$'

msg_unpause:
db 0d,'                                                                        ',0d,'$'

msg_abort:
db '   *** Abort ***',0d,0a,'$'

msg_looking:
db 'Searching drive '
msg_looking_1:
db 'A:',00

msg_attrib:
db '[     ]  $'

attr1:
db 'ADSHR'

msg_info_1:
db '     $'

msg_info_space:
db '  $'

msg_info_bytes:
db ' bytes$'

msg_info_dir:
db '<Dir>$'

msg_info_ampm:
db ' am  $'

msg_info_dash:
db '-$'

msg_info_colon:
db ':$'

msg_info_dot:
db '.$'

msg_info_open:
db ' ($'

msg_info_kb:
db ' KB)$'

msg_info_mb:
db ' MB)$'

msg_nothing:
db 'No matches found.',00

msg_num_entries:
db ' matches:  ',01,00

msg_num_files:
db ' files',01,00

msg_num_files1:
db ', ',01,00

msg_num_dirs:
db ' directories',01,00

msg_num_dirs1:
db '.',00

msg_hs:
db ' (',01,00

msg_hs1:
db ' H/S)',01,00

msg_maybe_more:
db '   /F$' 
msg_maybe_more1:
db ' was specified; there may be more.',00

msg_file_totals:
db '   Total of file sizes:  $'

msg_tree_max:
db 'Maximum value of TREE_END:  $'

star_dot_star:
db '*.*',00

msg_null  equ  star_dot_star + 3

var_locate:
db 'LOCATE='

var_path:
db 'PATH='

end_of_code:
end_code_seg  equ  (end_of_code / 10) + 21


;
;  ERRORLEVEL RETURNS:
;
;     00  File(s) found
;     01  No match found
;     02  Filespec error
;     03  User break may return this value
;     04  Memory allocation error
;     05  Bad DOS
;     06  Tree buffer overflow
;     10  Bad switch or syntax request
;
;
;  VERSION LOG:
;
;  v1.00   06-10-1995
;          Supports /T /X /R /N /D- and /D+
;
;  v1.01   06-12-1995
;          Fixes a bug in /T USE_PATH which prevented proper handling of the
;          root directory when it was current.  Does not attempt to search
;          drive B: on a single-drive system.
;
;  v1.02   06-15-1995
;          Provides a count of matches found.  The decimal-print routine
;          will probably crash if more than 650,000 matches are found.
;          No count messages will be printed if /N was specified.
;
;  v1.03   06-19-1995
;          Better decimal-print routine:  will correctly display up to
;          2^32-1 files.  If only one match is found, will display
;          "1 match" instead of "1 matches".
;
;  v1.04   06-22-1995
;          Hacked to bypass APPEND if installed and active.
;
;  v1.05   07-10-1995
;          Adds /H (list only hidden/system items.)  Final count is now
;          broken down into file and directory counts.
;
;  v1.06   10-21-1995
;          Adds /Fn (find first n files) and /I (displays more info,
;          in a somewhat Peter-like format.)
;  v1.06a  11-05-1995
;          The decimal-print routine now prints commas in large numbers.
;          Space bar allowed (as well as Return) to un-pause output.
;          Unused attributes shown as a period, instead of a space;
;          meanings of attribute letters are listed in the syntax message.
;  v1.06b  11-07-1995
;          Undocumented switch /$ suppresses display of drive letters.
;          Program will abort if TREE_END variable wraps around end of
;          64K buffer.
;  v1.06c  11-13-1995
;          Final file and directory counts include a subcount of hidden or
;          system files/directories.  If /H was specified, the subcounts
;          will be suppressed since they must match the main counts.
;
;  v1.07   11-20-1995
;          /$ behaves more sensibly (only searches one drive, invokes /N
;          automatically.)  If user specifies a filespec of only . or ..
;          these will now be handled correctly.  Parser is now simpler
;          (uses lookup-and-jump table, not monster case switch.)
;
;  v1.08   11-25-1995
;          Adds /B:"command" for automatic batch file creation.  If there
;          are no command line arguments at all, uses a default filespec
;          of *.*
;
;  v1.09   12-08-1995
;          Checks for duplicate drive mappings.  If more than one drive
;          letter is mapped to the same drive, only the lowest will be
;          searched.  /M defeats the drive-mapping check.
;  v1.09a  12-12-1995
;          New switch /S displays summary info (counts) only, not the
;          list of file/directory names.
;  v1.09b  12-14-1995
;          /R now officially skips network and SUBST'ed drives.
;  v1.09c  12-26-1995
;          Makes the pause routine more Peterlike.
;  v1.09d  12-28-1995
;          Allows a keystroke to pause a continuous listing.
;
;  v1.10   01-01-1996
;          Now checks for a LOCATE= environment variable.  All switches are
;          legal in the LOCATE variable except for /B.  Adds /F- which
;          'undoes' any previous /F, setting MAX_COUNT to zero.
;          New switches:  /W - wide Peter display.  /P - long Peter display.
;          /A - Peter display (long or wide).
;          Also fixes a bug in the handling of . and ..
;  v1.10a  01-03-1996
;          Trivial change in the handling of /P.  (Now identical to /A.)
;  v1.10b  01-17-1996
;          Allows use of the Control key to slow down scrolling.  Long live
;          the C64!
;
;  v1.11   02-02-1996
;          Fixes a bug:  LOCATE \FILENAME did not restrict the search to the
;          current drive, as documented; fixed.  Minimum DOS version reduced
;          to 3.2.  Error messages now printed to STDERR, not STDOUT.
;  v1.11a  02-06-1996
;          Long-Peter display now includes attributes.
;  v1.11b  02-09-1996
;          Includes a total of sizes of all files found.
;
;  v1.12   02-16-1996
;          Minor changes:  removes duplicate environment-search code in
;          USE_PATH; trivial changes to syntax display; removes an unused
;          error message.
;
